#include <fcntl.h>
#include <termios.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/ioctl.h>

#include "innovateSerial2.h"

int serialfd;
#define FALSE (0)
#define TRUE (1)
#define HEADERIDMASK	0xA280
#define HEADERID		0xA280
#define LC1IDMASK  0xE280
#define LC1ID	   0x4200


int serialInit(const char *devPath)
{struct termios term;

	if ((serialfd = open(devPath, O_RDWR)) < 0)
	{
		return -1;
	}
	
	// set up serial port
	
	memset(&term, 0, sizeof(term));
	tcgetattr(serialfd, &term);
	cfmakeraw(&term);
	term.c_iflag &= ~(ICRNL|INPCK);
	term.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);;
	
	term.c_cflag = (CS8 | CLOCAL);
	cfsetispeed(&term, B19200);
	cfsetospeed(&term, B19200);
	tcsetattr(serialfd, TCSAFLUSH, &term);
	
	return 1;
}

word readWord()
{word buffer=0;
 int numread, bytesavailable;
 
 	ioctl(serialfd, FIONREAD, &bytesavailable);
 	if (bytesavailable>=2) {
	 	numread=read(serialfd, &buffer, 2);
		if (numread == 2) {
			//printf("Got a word: %x\n",((buffer & 0x00FF)<<8) + ((buffer & 0xFF00)>>8));
			return ((buffer & 0x00FF)<<8) + ((buffer & 0xFF00)>>8);
		} else
			return 0;
	} else return 0;
}

int isHeader(word theWord, int *isSensorData, int *numWords)
{
	if ((theWord & HEADERIDMASK) == HEADERID) {
		if ((theWord & 0x1000) == 0x1000)
			*isSensorData=TRUE;
		*numWords=(int) ((theWord & 0x0100)>>1 | (theWord & 0x007F));
		return TRUE;
	} else {
		return FALSE;
	}
}

int isLC1Head(word theWord, int *status, int *AFRMult)
{
	if ((theWord & LC1IDMASK) == LC1ID) {
		*AFRMult = (int) (((theWord & 0x0100)>> 1) | (theWord & 0x007F));
		*status = (int) ((theWord & 0x1C00)>>10);
		return TRUE;
	} else return FALSE;
}

int getLC1Data(word theWord)
{
	return (int) (((theWord & 0x3F00)>> 1) | (theWord & 0x007F));
}

int readWB(int *lambda, float *AFR)
{word theWord;
 int status=IS2_NODATA;
 int isSensorData, numWords, wordCount, AFRMult;
 
	theWord=readWord();
	if (theWord == 0)
		return IS2_NODATA;
	else {
		if (isHeader(theWord, &isSensorData, &numWords) == TRUE) {
			//printf("Looks like a header\n");
			if (isSensorData==TRUE) {
				//printf("The word seems to contain sensor data in %d words\n", numWords);
				for (wordCount=0; wordCount<numWords; wordCount++) {
					theWord=readWord();
					if (isLC1Head(theWord, &status, &AFRMult)) {
						theWord=readWord();
						*lambda=getLC1Data(theWord);
						if (status==IS2_HAVELAMBDA) {
							*lambda+=500;
							*AFR=(float) *lambda*AFRMult/10000;
							
						} else *AFR=0.0;
						wordCount++;
					}
				}
			}
		} 
		
	}
	return status;
}

void serialClose()
{
	close(serialfd);

}
