#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/timeb.h>
#include <unistd.h>
#include <math.h>

#include <asm/arch-sa1100/hijack.h>

#define FALSE (0)
#define TRUE  (1)
#define STOICH 14.7

#include "vfdlib.h"
#include "innovateSerial2.h"

#define SERIALDEVICE "/dev/ttyS1"

struct tickStruct {
	int x,x1,y,y1;
	char val[5];
};

unsigned char screenbuf[EMPEG_SCREEN_BYTES] = {0,};
const char *hijack_menu[] = { "Wideband O2", NULL };
int displayfd = -1;
unsigned    buttons[9] = {8,
			IR_LEFT_BUTTON_PRESSED,
			IR_RIGHT_BUTTON_PRESSED,
			IR_TOP_BUTTON_PRESSED,
			IR_BOTTOM_BUTTON_PRESSED,
			IR_KNOB_PRESSED,
			IR_KNOB_RIGHT,
			IR_KNOB_LEFT
			};
float afrBuf[EMPEG_SCREEN_COLS] = {0,};
int lambdaBuf[EMPEG_SCREEN_COLS] = {0,};
int lambdaBufPos=0;
int afrBufPos=0;
struct timeb prevSampleTime, currentTime;
int sampleInterval;
int notConnectedCount=0,prevStatus=IS2_NODATA;


hijack_geom_t geom = {8,24,32,100}, fullscreen = {0,EMPEG_SCREEN_ROWS-1,0,EMPEG_SCREEN_COLS-1};

int compareTime(struct timeb *curr, struct timeb *prev, int interval)
{struct timeb elapsed;
 
 long int tempInt;
 
 	elapsed.time=curr->time-prev->time;
 	if (curr->millitm >= prev->millitm)
 		elapsed.millitm=curr->millitm-prev->millitm;
 	else {
 		elapsed.time--;
 		elapsed.millitm=1000+elapsed.millitm-prev->millitm;
 	}
 	
 	tempInt=elapsed.time*1000+elapsed.millitm;

	if (tempInt>interval)
 		return TRUE;
	else
 		return FALSE;
}


void drawAFR(float afr, int x, int y)
{
	char tempString[5];
	int legendWidth;
	legendWidth=vfdlib_getTextWidth("Air/Fuel Ratio", 0);
	vfdlib_drawText(screenbuf, "Air/Fuel Ratio", x+2, y, 0, VFDSHADE_MEDIUM);
	vfdlib_drawOutlineRectangleClipped(screenbuf, x, y+vfdlib_getTextHeight(0)+2, x+legendWidth+4, y+vfdlib_getTextHeight(0)+vfdlib_getTextHeight(2)+2, VFDSHADE_MEDIUM);
	sprintf(tempString,"%.2f",afr);
	vfdlib_drawText(screenbuf, tempString, x, y+vfdlib_getTextHeight(0)+4, 2, VFDSHADE_BRIGHT);
	
}

void drawLambda(float lambda, int x, int y)
{
	char tempString[5];
	int legendWidth;
	legendWidth=vfdlib_getTextWidth("Lambda", 0);
	vfdlib_drawText(screenbuf, "Lambda", x+3, y, 0, VFDSHADE_MEDIUM);
	vfdlib_drawOutlineRectangleClipped(screenbuf, x, y+vfdlib_getTextHeight(0)+2, x+legendWidth+6, y+vfdlib_getTextHeight(0)+vfdlib_getTextHeight(2)+2, VFDSHADE_MEDIUM);
	sprintf(tempString,"%.2f",lambda);
	vfdlib_drawText(screenbuf, tempString, x, y+vfdlib_getTextHeight(0)+4, 2, VFDSHADE_BRIGHT);
	
}

void drawAFRCompact(float afr, int x, int y, int font)
{char tempString[5];
	
	sprintf(tempString, "%.2f", afr);
	vfdlib_drawText(screenbuf, "AFR:", x, y, font, VFDSHADE_MEDIUM);
	vfdlib_drawText(screenbuf, tempString, x+vfdlib_getTextWidth("AFR: ", font), y, font, VFDSHADE_BRIGHT);
}

void drawLambdaCompact(float lambda, int x, int y, int font)
{char tempString[5];
	
	sprintf(tempString, "%.2f", lambda);
	vfdlib_drawText(screenbuf, "Lambda:", x, y, font, VFDSHADE_MEDIUM);
	vfdlib_drawText(screenbuf, tempString, x+vfdlib_getTextWidth("Lambda: ", font), y, font, VFDSHADE_BRIGHT);
}

int afrToY(int height, int center, int range, float afr)
{int yval;
	yval=height/2+(int)(((afr-center)/(float)range)*(float)height);
	return yval;
}

int lambdaToY(int height, int center, int range, int lambda)
{int yval;
	yval=height/2+(int)((lambda-center)/range * height);
	return yval;
}


void drawGraph(int height, int center, int range)
{int pos, xval=0;
 
	
	for (pos=afrBufPos; pos<EMPEG_SCREEN_COLS; pos++) {
		vfdlib_drawPointClipped(screenbuf,xval,afrToY(height,center,range,afrBuf[pos]),VFDSHADE_BRIGHT);
		xval++;
	}
	for (pos=0; pos<afrBufPos; pos++) {
		vfdlib_drawPointClipped(screenbuf,xval,afrToY(height,center,range,afrBuf[pos]),VFDSHADE_BRIGHT);
		xval++;
	}
	
	vfdlib_drawLineHorizClipped(screenbuf, afrToY(height,center,range,STOICH), 0,
		EMPEG_SCREEN_COLS, VFDSHADE_DIM);
}

void drawLambdaGraph(int height, int center, int range)
{int pos, xval=0;

	for (pos=lambdaBufPos; pos>EMPEG_SCREEN_COLS; pos++) {
		vfdlib_drawPointClipped(screenbuf,xval,lambdaToY(height,center,range,lambdaBuf[pos]),VFDSHADE_BRIGHT);
		xval++;
	}
	for (pos=0; pos<lambdaBufPos; pos++) {
		vfdlib_drawPointClipped(screenbuf,xval,lambdaToY(height,center,range,lambdaBuf[pos]),VFDSHADE_BRIGHT);
		xval++;
	}
	vfdlib_drawLineHorizClipped(screenbuf,lambdaToY(height,center,range,1000), 0,
		EMPEG_SCREEN_COLS,VFDSHADE_DIM);
	

}

double lambdaToAngle(float lambda)
{
	 return (lambda-0.5)*M_PI-M_PI_2;
}


void drawGauge(int centerX, float lambda)
{double angle=0;
 float l;
	int x, y, x1, y1, radius=EMPEG_SCREEN_ROWS-5, centerY=EMPEG_SCREEN_ROWS;
	char text[10];

	
	vfdlib_drawOutlineEllipseClipped(screenbuf, centerX, centerY, radius, radius, VFDSHADE_MEDIUM);
	for (l=0.6; l<1.5; l+=0.1) {
		sprintf(text,"%.1f",l);
		angle=lambdaToAngle(l);
		x=centerX+(int)((radius) * sin(angle));
		y=centerY-(int)((radius) * cos(angle));
		x1=centerX+(int)((radius-2) * sin(angle));
		y1=centerY-(int)((radius-2) * cos(angle));
		vfdlib_drawLineClipped(screenbuf, x, y, x1, y1,VFDSHADE_BRIGHT);
		if (x<centerX) {
			x=x-vfdlib_getTextWidth(text,0);
			y=y-vfdlib_getTextHeight(0)+1;
		} else if (x==centerX) {
			x=x-vfdlib_getTextWidth(text,0)/2;
			y=y-vfdlib_getTextHeight(0);
		} else {
			y=y-vfdlib_getTextHeight(0)+1;
		}
		vfdlib_drawText(screenbuf,text, x, y, 0, VFDSHADE_BRIGHT); 
	
	}
	
	if (lambda>=.5 && lambda <=1.5) {
		angle=lambdaToAngle(lambda);
		x=centerX+(int)((radius-3) * sin(angle));
		y=centerY-(int)((radius-3) * cos(angle));
		vfdlib_drawLineClipped(screenbuf, centerX, centerY, x, y, VFDSHADE_BRIGHT);
	}

	vfdlib_drawSolidEllipseClipped(screenbuf, centerX, centerY, 5, 5, VFDSHADE_MEDIUM);
	vfdlib_drawOutlineEllipseClipped(screenbuf, centerX, centerY, 5, 5, VFDSHADE_BRIGHT);
	vfdlib_drawOutlineEllipseClipped(screenbuf, centerX, centerY, 6, 6, VFDSHADE_BRIGHT);
	vfdlib_drawText(screenbuf, "RICH", centerX-radius-vfdlib_getTextWidth("RICH",0),centerY-vfdlib_getTextHeight(0),0,VFDSHADE_BRIGHT);
	vfdlib_drawText(screenbuf, "LEAN", centerX+radius,centerY-vfdlib_getTextHeight(0),0,VFDSHADE_BRIGHT);
	
}

void drawFastGauge(int centerX, float lambda)
{double angle=0;
 float l;
	int x, y, x1, y1, radius=EMPEG_SCREEN_ROWS-5, centerY=EMPEG_SCREEN_ROWS, count=0;
	static struct tickStruct ticks[11]={{0,0,0,0,""},};
 	
	if (ticks[0].x==0) {			// pre calculate the locations of the tick marks
		for (l=0.6; l<1.5; l+=0.1) {
			sprintf(ticks[count].val,"%.1f",l);
			angle=lambdaToAngle(l);
			ticks[count].x=(int)((radius) * sin(angle));
			ticks[count].y=(int)((radius) * cos(angle));
			ticks[count].x1=(int)((radius-2) * sin(angle));
			ticks[count].y1=(int)((radius-2) * cos(angle));
		}
	}
	
	vfdlib_drawOutlineEllipseClipped(screenbuf, centerX, centerY, radius, radius, VFDSHADE_MEDIUM);
	count=0;
	while (ticks[count].x!=0) {
		x=centerX+ticks[count].x;
		x1=centerX+ticks[count].x1;
		y=centerY-ticks[count].y;
		y1=centerY-ticks[count].y1;
		
		vfdlib_drawLineClipped(screenbuf, x, y, x1, y1, VFDSHADE_BRIGHT);	
		if (x<centerX) {
			x=x-vfdlib_getTextWidth(ticks[count].val,0);
			y=y-vfdlib_getTextHeight(0)+1;
		} else if (x==centerX) {
			x=x-vfdlib_getTextWidth(ticks[count].val,0)/2;
			y=y-vfdlib_getTextHeight(0);
		} else {
			y=y-vfdlib_getTextHeight(0)+1;
		}
		vfdlib_drawText(screenbuf,ticks[count].val, x, y, 0, VFDSHADE_BRIGHT); 
	}
	
	if (lambda>=.5 && lambda <=1.5) {
		angle=lambdaToAngle(lambda);
		x=centerX+(int)((radius-3) * sin(angle));
		y=centerY-(int)((radius-3) * cos(angle));
		vfdlib_drawLineClipped(screenbuf, centerX, centerY, x, y, VFDSHADE_BRIGHT);
	}

	vfdlib_drawSolidEllipseClipped(screenbuf, centerX, centerY, 5, 5, VFDSHADE_MEDIUM);
	vfdlib_drawOutlineEllipseClipped(screenbuf, centerX, centerY, 5, 5, VFDSHADE_BRIGHT);
	vfdlib_drawOutlineEllipseClipped(screenbuf, centerX, centerY, 6, 6, VFDSHADE_BRIGHT);
	vfdlib_drawText(screenbuf, "RICH", centerX-radius-vfdlib_getTextWidth("RICH",0),centerY-vfdlib_getTextHeight(0),0,VFDSHADE_BRIGHT);
	vfdlib_drawText(screenbuf, "LEAN", centerX+radius,centerY-vfdlib_getTextHeight(0),0,VFDSHADE_BRIGHT);
	
}


void drawStatus(int x, int y, int status, int lambda)
{char tempString[255];
	switch (status) {
		case IS2_NODATA:
			vfdlib_drawSolidRectangleClipped(screenbuf, x, y, EMPEG_SCREEN_COLS, y+vfdlib_getTextHeight(0)+2, VFDSHADE_BLACK);
			vfdlib_drawText(screenbuf, "Not Connected", x, y, 0, VFDSHADE_BRIGHT);
			break;
		case IS2_HAVEO2PERCENT:
			vfdlib_drawSolidRectangleClipped(screenbuf, x, y, EMPEG_SCREEN_COLS, y+vfdlib_getTextHeight(0)*2+2, VFDSHADE_BLACK);
			vfdlib_drawText(screenbuf, "Out of range", x, y , 0, VFDSHADE_BRIGHT);
			sprintf(tempString,"%.1f%% oxygen", lambda/10.0);
			vfdlib_drawText(screenbuf, tempString, x, vfdlib_getTextHeight(0)+2, 0, VFDSHADE_BRIGHT);
			break;
		case IS2_FREEAIRCAL:
			vfdlib_drawSolidRectangleClipped(screenbuf, x, y, EMPEG_SCREEN_COLS, y+vfdlib_getTextHeight(0)+2, VFDSHADE_BLACK);
			vfdlib_drawText(screenbuf, "Free air calibration in progress", x, y, 0, VFDSHADE_BRIGHT);
			break;
		case IS2_NEEDCALIBRATION:
			vfdlib_drawSolidRectangleClipped(screenbuf, x, y, EMPEG_SCREEN_COLS, y+vfdlib_getTextHeight(0)+2, VFDSHADE_BLACK);
			vfdlib_drawText(screenbuf, "In need of free air calibration", x, y, 0, VFDSHADE_BRIGHT);
			break;
		case IS2_WARMINGUP:
			vfdlib_drawSolidRectangleClipped(screenbuf, x, y, EMPEG_SCREEN_COLS, y+vfdlib_getTextHeight(0)*2+2, VFDSHADE_BLACK);
			vfdlib_drawText(screenbuf, "Sensor warming up", x, y, 0, VFDSHADE_BRIGHT);
			sprintf(tempString,"at %d%% of operating temperature", lambda/10);
			vfdlib_drawText(screenbuf, tempString, x, vfdlib_getTextHeight(0)+2, 0, VFDSHADE_BRIGHT);
			break;
		case IS2_HEATERCAL:
			vfdlib_drawSolidRectangleClipped(screenbuf, x, y, EMPEG_SCREEN_COLS, y+vfdlib_getTextHeight(0)*2+2, VFDSHADE_BLACK);
			vfdlib_drawText(screenbuf, "Heater calibration in progress", x, y, 0, VFDSHADE_BRIGHT);
			sprintf(tempString,"%d", lambda);
			vfdlib_drawText(screenbuf, tempString, x, vfdlib_getTextHeight(0)+2, 0, VFDSHADE_BRIGHT);
			break;
		case IS2_ERROR:
			vfdlib_drawSolidRectangleClipped(screenbuf, x, y, EMPEG_SCREEN_COLS, y+vfdlib_getTextHeight(0)*2+2, VFDSHADE_BLACK);
			vfdlib_drawText(screenbuf, "Wideband Error", x, y, 0, VFDSHADE_BRIGHT);
			switch (lambda) {
				case 1:
					sprintf(tempString,"#1: Short in heater circuit");
					break;
				case 2:
					sprintf(tempString,"#2: Heater circuit open (no sensor?)");
					break;
				case 3:
					sprintf(tempString,"#3: Pump circuit shorted");
					break;
				case 4:
					sprintf(tempString,"#4: Pump circuit open");
					break;
				case 5:
					sprintf(tempString,"#5: Reference circuit shorted");
					break;
				case 6:
					sprintf(tempString,"#6: Reference circuit open");
					break;
				case 7:
					sprintf(tempString,"#7: General software error");
					break;
				case 8:
					sprintf(tempString,"#8: Sensor timing error (too hot?)");
					break;
				case 9:
					sprintf(tempString,"#9: Supply voltage too low");
					break;
			}
			vfdlib_drawText(screenbuf, tempString, x, vfdlib_getTextHeight(0)+2, 0, VFDSHADE_BRIGHT);
			break;
	}	
}

void updateDisplay()
{
	if (ioctl(displayfd, EMPEG_HIJACK_DISPWRITE, screenbuf) == -1) {
    	fprintf(stderr, "Unable to update display.\n");
    	exit(1);
	}
}

int main(int argc, char **argv)
{
	int done = FALSE, quit=FALSE;
	int f=0, status, lambda=0;
	int viewMode=2;
	float afr=0;
	unsigned int buttonData;
	
	if ((displayfd = open("/dev/display", O_RDWR)) == -1) {
		fprintf(stderr, "Unable to open display.\n");
		exit(1);
	}
	
	while (!quit) {
		if (ioctl(displayfd, EMPEG_HIJACK_WAITMENU, hijack_menu) == -1) {
			fprintf(stderr, "Unable to bind into Hijack menu.\n");
			return 1;
		}
		
		sampleInterval=50;
		
		prevSampleTime.time=0;
		prevSampleTime.millitm=0;
		
		vfdlib_registerFont("/empeg/lib/fonts/small.bf", 0);
		vfdlib_registerFont("/empeg/lib/fonts/medium.bf", 1);
		vfdlib_registerFont("/empeg/lib/fonts/large.bf", 2);
		
		if (ioctl(displayfd, EMPEG_HIJACK_BINDBUTTONS, buttons) == -1) {
			fprintf(stderr, "Unable to bind buttons.\n");
			return 1;
		}
		
		if (ioctl(displayfd,EMPEG_HIJACK_SETGEOM, &fullscreen) == -1) {
			fprintf(stderr, "Unable to set geometry\n");
			return 1;
		}
		
	
		if (serialInit(SERIALDEVICE) == -1) {
			fprintf(stderr, "Unable to open serial device\n");
			return 1;
		}
		
		status=readWB(&lambda, &afr);
		
		done = FALSE;
		while (!done) {
			if (ioctl(displayfd, EMPEG_HIJACK_POLLBUTTONS, &buttonData) != -1) {
				switch (buttonData) {
					case IR_RIGHT_BUTTON_PRESSED:
						break;
					case IR_LEFT_BUTTON_PRESSED:
						break;
					case IR_BOTTOM_BUTTON_PRESSED:
						if (viewMode<2)
							viewMode++;
						else
							viewMode=0;
						break;
					case IR_KNOB_PRESSED:
						done=TRUE;
						quit=TRUE;
						break;
					case IR_KNOB_RIGHT:
						if (sampleInterval < 3000)
							sampleInterval += 25;
						break;
					case IR_KNOB_LEFT:
						if (sampleInterval > 25)
							sampleInterval -= 25;
						break;
					case IR_TOP_BUTTON_PRESSED:
						done = TRUE;
						break;
				}
			}
			status=readWB(&lambda, &afr);
			if (status==IS2_NODATA)
				notConnectedCount++;
			else
				notConnectedCount=0;
			if (notConnectedCount<100 && status==IS2_NODATA)
				status=prevStatus;
			ftime(&currentTime);
			if ((status==IS2_HAVELAMBDA) && (compareTime(&currentTime, &prevSampleTime, sampleInterval))) {
				prevSampleTime=currentTime;
				lambdaBuf[lambdaBufPos]=lambda;
				lambdaBufPos++;
				if (lambdaBufPos>EMPEG_SCREEN_COLS)
					lambdaBufPos=0;
				// afrBuf[afrBufPos]=afr;
				// afrBufPos++;
				// if (afrBufPos>EMPEG_SCREEN_COLS)
				//	afrBufPos=0;
				
			}
			prevStatus=status;
			
			vfdlib_fastclear(screenbuf);
			switch (viewMode) {
				case 0:
					drawAFRCompact(afr, 0, EMPEG_SCREEN_ROWS-vfdlib_getTextHeight(f), 0);
					drawLambdaCompact(lambda/1000.0, EMPEG_SCREEN_COLS/2, EMPEG_SCREEN_ROWS-vfdlib_getTextHeight(f), 0);
					drawLambdaGraph(EMPEG_SCREEN_ROWS-vfdlib_getTextHeight(f), 1000, 1000);
					drawStatus(0,0,status,lambda);
					break;
				case 1:
					drawAFR(afr, 0, 0);
					drawLambda(lambda/1000.0, EMPEG_SCREEN_COLS/2, 0);
					drawStatus(0,0,status,lambda);
					break;
				case 2:
					drawGauge(EMPEG_SCREEN_COLS-EMPEG_SCREEN_ROWS-12,lambda/1000.0);
					drawLambda(lambda/1000.0, 10,0);
					drawStatus(0,0,status,lambda);
					break;
			}
			
			updateDisplay();
			prevStatus=status;
			if (quit==TRUE) {
				vfdlib_fastclear(screenbuf);
				vfdlib_drawText(screenbuf, "Are you sure you want to quit?", 2, 2, 1, VFDSHADE_BRIGHT);
				vfdlib_drawText(screenbuf, "To quit press knob, any other button cancels", 5, vfdlib_getTextHeight(1)+2, 1, VFDSHADE_BRIGHT);
				while (ioctl(displayfd, EMPEG_HIJACK_POLLBUTTONS, &buttonData) != -1)
				{
				}
				if (buttonData!=IR_KNOB_PRESSED) {
					quit=FALSE;
					done=FALSE;
				}
			}
		}
		
		vfdlib_unregisterAllFonts();
		serialClose();
	}
	
	
	
	if (ioctl(displayfd,EMPEG_HIJACK_WAITMENU, NULL) == -1) {
		fprintf(stderr, "Unable to release menu system.\n");
		return 1;
	}

	return 0;
}

